package kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011;

import junit.framework.TestCase;

import kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011.SymbolicHsviIPPC2011;
import kr.ac.kaist.pomdp.symbolic.perseus.SymbolicProblem;

/**
 * junit test case for SymbolicHsviIPPC2011
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class SymbolicHsviIPPC2011Test extends TestCase {

	public void testRun() throws Exception
	{
		// How to use autoRun() function
		// Parameters
		// 1: POMDP SYM file directory
		// 2: RDDL problem name
		// 3: RDDL instance name
		// 4: Goal width
		// 5: precision ratio
		// 6: # initialize horizon
		// 7: Whether to use masking
		// 8: Whether to use symmetry
		// 9: Whether to simulate periodically
		// 10: # of periodical simulation runs
		// 11: policy savefile directory
		// 12: permutation savefile directory
		
		String probDir = "./data/final_comp/spudd_sperseus/";
		String probName = "game_of_life";
		String probInstName = "game_of_life_inst_pomdp__1";
		double goalwidth = 0.1;
		double precRatio = 0.95;
		int nInitHorizon = Integer.MAX_VALUE;
		boolean bUseMasking	= true;
		boolean bUseSymmetry = true;
		boolean bPeSim = true;
		int nPeSimNum = 30;
		String policyDir = "./policy/";
		String permDir = "./permutation/";
		
		// if there is the class file for the problem, then read the file and continue the class 
		autoRun(probDir, probName, probInstName, goalwidth, precRatio, nInitHorizon, bUseMasking, 
				bUseSymmetry, bPeSim, nPeSimNum, policyDir, permDir);
		
	}

	public static void autoRun(String probDir, String rddlName, String instanceName, double goalwidth, double precRatio, int initHorizon, 
			boolean bUseMasking, boolean bUseSymmetry, boolean bPeSim, int nPeSimNum,	String policyDir, String permDir) throws Exception
	{
		// read POMDP specification file
		String pomdpFilename = null;
		SymbolicProblem symPomdp = null;
		pomdpFilename = probDir + instanceName + ".sperseus";
		symPomdp = new SymbolicProblem(pomdpFilename, false);

		// build SymbolicHsviDdTreeFiniteSymmetry instance
		SymbolicHsviIPPC2011 solver = null;
		String className = "symhsviddtreefinitesym";
		String policyFilename = policyDir + className + "_" + instanceName + ".policy";
		String permFilename = permDir + instanceName + ".perms";
			
		solver = new SymbolicHsviIPPC2011(symPomdp, goalwidth, initHorizon, bUseMasking, bUseSymmetry, policyFilename, permFilename, bPeSim, nPeSimNum);
		
		// run SymbolicHsvi algorithm
		solver.run(precRatio);
	}
	
}
