package kr.ac.kaist.pomdp.data;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import kr.ac.kaist.pomdp.symbolic.perseus.DD;

/**
 * policy wrapping class
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class PolicyMaintainer {

	public static void saveAlphas(java.util.Vector<DD> alphas,
			java.util.Vector<Integer> policy, java.util.Vector<DD[]> masks,
			String filename) {
		try {
			FileOutputStream fw = new FileOutputStream(filename);
			ObjectOutputStream bw = new ObjectOutputStream(fw);

			// flags
			bw.writeObject(true); // use alpha vector masking

			// policy save routine
			for (int i = 0; i < alphas.size(); i++) {
				DD cAlpha = alphas.get(i);
				Integer cPolicy = policy.get(i);
				DD[] mask = masks.get(i);

				bw.writeObject(cPolicy);
				bw.writeObject(cAlpha);
				bw.writeObject(mask);
			}

			bw.close();
			fw.close();
		} catch (Exception e) {
		}
	}

	public static void saveAlphas(java.util.Vector<DD> alphas,
			java.util.Vector<Integer> policy, String filename) {
		try {
			FileOutputStream fw = new FileOutputStream(filename);
			ObjectOutputStream bw = new ObjectOutputStream(fw);

			// flags
			bw.writeObject(false); // don't use alpha vector masking

			// policy save routine
			for (int i = 0; i < alphas.size(); i++) {
				DD cAlpha = alphas.get(i);
				Integer cPolicy = policy.get(i);
				bw.writeObject(cPolicy);
				bw.writeObject(cAlpha);
			}

			bw.close();
			fw.close();
		} catch (Exception e) {
		}
	}

	public static void saveAlphasFinite(java.util.Vector<DD>[] alphas,
			java.util.Vector<Integer>[] policy, String filename) {
		try {
			FileOutputStream fw = new FileOutputStream(filename);
			ObjectOutputStream bw = new ObjectOutputStream(fw);

			bw.writeObject(alphas.length);
			for (int h = 0; h < alphas.length; h++)
				bw.writeObject(alphas[h].size());

			for (int h = 0; h < alphas.length; h++) {
				for (int i = 0; i < alphas[h].size(); i++) {
					DD cAlpha = alphas[h].get(i);
					Integer cPolicy = policy[h].get(i);
					bw.writeObject(cPolicy);
					bw.writeObject(cAlpha);
				}
			}

			bw.close();
			fw.close();
		} catch (Exception e) {
		}
	}

	public static ArrayList loadAlphasDDFinite(String filename) {
		ArrayList result = new ArrayList();
		java.util.Vector<DD>[] alphas = null;
		java.util.Vector<Integer>[] policy = null;

		try {
			FileInputStream fi = new FileInputStream(filename);
			ObjectInputStream ois = new ObjectInputStream(fi);

			int horizon = (Integer) ois.readObject();
			alphas = new java.util.Vector[horizon];
			policy = new java.util.Vector[horizon];
			int[] num = new int[horizon];
			for (int h = 0; h < horizon; h++) {
				num[h] = (Integer) ois.readObject();
			}

			for (int h = 0; h < horizon; h++) {
				for (int i = 0; i < num[h]; i++) {
					Integer cPolicy = (Integer) ois.readObject();
					DD cAlpha = (DD) ois.readObject();
					policy[h].add(cPolicy);
					alphas[h].add(cAlpha);
				}
			}
		} catch (Exception e) {
		}

		result.add(alphas);
		result.add(policy);
		return result;
	}

	public static ArrayList loadAlphasDD(String filename) {
		ArrayList result = new ArrayList();
		java.util.Vector<DD> alphas = new java.util.Vector();
		java.util.Vector<Integer> policy = new java.util.Vector();
		java.util.Vector<DD[]> masks = new java.util.Vector();

		boolean useMasking = false;

		try {
			FileInputStream fi = new FileInputStream(filename);
			ObjectInputStream ois = new ObjectInputStream(fi);

			useMasking = (Boolean) ois.readObject();

			while (true) {
				Integer cPolicy = (Integer) ois.readObject();
				if (cPolicy == null)
					break;
				DD cAlpha = (DD) ois.readObject();
				policy.add(cPolicy);
				alphas.add(cAlpha);

				if (useMasking) {
					DD[] mask = (DD[]) ois.readObject();
					masks.add(mask);
				}
			}
		} catch (Exception e) {
		}

		result.add(alphas);
		result.add(policy);
		if (useMasking)
			result.add(masks);
		return result;
	}

}