package kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011;

import java.io.PrintStream;

import kr.ac.kaist.pomdp.symbolic.perseus.Config;

/**
 * BeliefDdLeaf class - leaf node in BeliefDd
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class EdgeDdLeaf extends EdgeDd {
	private static final long serialVersionUID = 3246082536994495972L;

	private Edge edge;
	private int[][] config; // configuration
	// public double[] obsVals;
	// public double maxNextValue;
	public Alpha obsStrat;

	/**
	 * DDleaf constructor
	 * 
	 * @param val
	 *            input value
	 */
	private EdgeDdLeaf(Edge edge) {
		this.edge = edge;
		this.var = 0;
		this.config = null;
	}

	/**
	 * DDleaf constructor
	 * 
	 * @param val
	 *            input value
	 * @param config
	 *            input configuration
	 */
	private EdgeDdLeaf(Edge edge, int[][] config) {
		this.edge = edge;
		this.var = 0;
		this.config = config;
	}

	/**
	 * get new DDleaf instance using leafHashtable
	 * 
	 * @param val
	 *            input value
	 * @return new DDleaf instance
	 */
	public static EdgeDd myNew(Edge edge) {

		// create new leaf
		EdgeDdLeaf leaf = new EdgeDdLeaf(edge);

		// try to lookup leaf in leafHashtable
		// WeakReference storedLeaf = (WeakReference)
		// Global.leafHashtable.get(leaf);
		// if (storedLeaf != null)
		// return (BeliefDdLeaf) storedLeaf.get();

		// store leaf in leafHashtable
		// Global.leafHashtable.put(leaf, new WeakReference<DD>(leaf));
		return leaf;
	}

	/**
	 * get new DDleaf instance using leafHashtable
	 * 
	 * @param val
	 *            input value
	 * @param config
	 *            input configuration
	 * @return new DDleaf instance
	 */
	public static EdgeDd myNew(Edge edge, int[][] config) {

		// create new leaf
		EdgeDdLeaf leaf = new EdgeDdLeaf(edge, config);

		// try to lookup leaf in leafHashtable
		// WeakReference storedLeaf = (WeakReference)
		// Global.leafHashtable.get(leaf);
		// if (storedLeaf != null)
		// return (EdgeDdLeaf) storedLeaf.get();

		// store leaf in leafHashtable
		// Global.leafHashtable.put(leaf, new WeakReference<DD>(leaf));
		return leaf;
	}

	/**
	 * get variable set there is no information about variables in a leaf node.
	 */
	public int[] getVarSet() {
		return new int[0];
	}

	/**
	 * get value
	 */
	public Edge getEdge() {
		return edge;
	}

	public String debugEdgeConfig() {
		String temp = "";
		for (int i = 0; i < edge.obsConfig.length; i++) {
			temp += "obs" + i + ": " + edge.obsConfig[i] + " ";
		}
		temp += "\n";
		return temp;
	}

	/**
	 * get configuration
	 */
	public int[][] getConfig() {
		return config;
	}

	/**
	 * get the number of leaves of this node always return 1
	 */
	public int getNumLeaves() {
		return 1;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {

		if (obj.getClass() != getClass())
			return false;
		EdgeDdLeaf leaf = (EdgeDdLeaf) obj;

		if (edge.equals(leaf.edge) && Config.equals(config, leaf.config))
			return true;
		else
			return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return edge.hashCode() + Config.hashCode(config);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kr.ac.kaist.pomdp.symbolic.DD#store()
	 */
	public EdgeDd store() {
		return EdgeDdLeaf.myNew(edge, config);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kr.ac.kaist.pomdp.symbolic.DD#display(java.lang.String)
	 */
	public void display(String space) {
		// System.out
		// .println(space + "leaf: " + Double.toString(val) + "  " +
		// Config.toString(config));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kr.ac.kaist.pomdp.symbolic.DD#display(java.lang.String,
	 * java.lang.String)
	 */
	public void display(String space, String prefix) {
		// System.out.println(space + prefix + Double.toString(val) + "  " +
		// Config.toString(config));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kr.ac.kaist.pomdp.symbolic.DD#printSpuddDD(java.io.PrintStream)
	 */
	public void printSpuddDD(PrintStream ps) {
		// ps.print("(" + Double.toString(val) + ")");
	}
}
