package kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011;

import java.io.Serializable;

import kr.ac.kaist.pomdp.symbolic.perseus.DD;
import kr.ac.kaist.pomdp.symbolic.perseus.OP;

/**
 * Node class for a POMDP tree structure (public member "horizon" is added for
 * finite-horizon problems)
 * 
 * @see T. Smith and R. Simmons, "Heuristic search value iteration for POMDPs,"
 *      UAI-2004.
 * @see T. Smith and R. Simmons,
 *      "Point-based POMDP algorithms: Improved analysis and implementation,"
 *      UAI-2005.
 * @see H. S. Sim et al.,
 *      "Symbolic heuristic search value iteration for factored POMDPs,"
 *      AAAI-2008.
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 */
public class Node implements Serializable {
	private static final long serialVersionUID = 6953196082975358534L;

	public int horizon;

	public DD[] b; // belief state
	public int filled; // # of non-zero entries of b
	public QEntry[] Q;
	public double lbVal;
	public double ubVal;
	public MaxPlaneData bdata;
	public int[] stateVars;

	public Node(int horizon) {
		this.horizon = horizon;
		lbVal = Double.NEGATIVE_INFINITY;
		ubVal = Double.NEGATIVE_INFINITY;
	}

	/**
	 * Node constructor
	 */
	public Node(DD[] b, int[] stateVarsId, int horizon) {
		this.horizon = horizon;
		this.b = b;
		this.stateVars = stateVarsId;
		lbVal = Double.NEGATIVE_INFINITY;
		ubVal = Double.NEGATIVE_INFINITY;
	}

	public int hashCode() {
		double sum = 0.0;
		for (int i = 0; i < b.length; i++) {
			sum += OP.dotProductNoMem(b[i], b[i], stateVars);
		}
		return ((Double) sum).hashCode();
	}

	public boolean equals(Object o) {

		if (o == null)
			return false;

		if (horizon != ((Node) o).horizon)
			return false;

		DD[] target = ((Node) o).b;
		for (int i = 0; i < target.length; i++) {
			if (!OP.maxNormDiff(this.b[i], target[i], 1e-6))
				return false;
		}
		return true;
	}
}
