package kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011;

import java.io.*;
import java.util.ArrayList;

/**
 * PermGenerator Class for computing all permutations using the given generators
 * 
 * usage 1. SymbolicSymmetry create graph file of a problem ex)
 * "game_of_life_inst_pomdp__1.sperseus.pomdpGraph" 2. bliss find graph
 * automorphism using bliss software ex)
 * "./bliss game_of_life_inst_pomdp__1.sperseus.pomdpGraph" 3. PermGenerator
 * From the result of bliss, input the generators of bliss, and find all
 * permutations
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class PermGenerator {

	public static String dir = "./permutation/";
	public static String instName = "game_of_life_inst_pomdp_1";

	public static int numVars = 9 * 2;
	public static int numActions = 10;

	public static int gensVar[][][] = new int[][][] {
			{ { 3, 5 }, { 6, 8 }, { 9, 18 }, { 10, 17 }, { 11, 16 },
					{ 12, 14 }, { 13, 15 }, { 21, 26 }, { 22, 27 }, { 23, 25 } },
			{ { 2, 3 } } };
	public static int gensAct[][][] = new int[][][] {
			{ { 1, 3 }, { 4, 6 }, { 7, 9 } }, { { 2, 4 }, { 3, 7 }, { 6, 8 } } };

	public static void main(String[] args) throws Exception {

		String filename = dir + instName + ".perms";

		ArrayList<int[]> varPerms = new ArrayList<int[]>();
		ArrayList<int[]> actPerms = new ArrayList<int[]>();
		int[] idenVar = new int[numVars];
		for (int i = 0; i < numVars; i++)
			idenVar[i] = i + 1;
		varPerms.add(idenVar);
		int[] idenAct = new int[numActions];
		for (int i = 0; i < numActions; i++)
			idenAct[i] = i;
		actPerms.add(idenAct);

		boolean bChange = true;
		while (bChange) {
			bChange = false;

			// printPerms(varPerms, actPerms);

			int nPerms = varPerms.size();
			for (int perm = 0; perm < nPerms; perm++) {
				int[] varPerm = varPerms.get(perm);
				int[] actPerm = actPerms.get(perm);

				for (int i = 0; i < gensVar.length; i++) {
					int[] newVarPerm = varPerm.clone();
					int[] newActPerm = actPerm.clone();

					for (int j = 0; j < gensVar[i].length; j++) {
						int leng = gensVar[i][j].length;
						for (int k = 0; k < gensVar[i][j].length; k++) {
							newVarPerm[gensVar[i][j][k] - 1] = varPerm[gensVar[i][j][(k + 1)
									% leng] - 1];
						}
					}

					for (int j = 0; j < gensAct[i].length; j++) {
						int leng = gensAct[i][j].length;
						for (int k = 0; k < gensAct[i][j].length; k++) {
							newActPerm[gensAct[i][j][k]] = actPerm[gensAct[i][j][(k + 1)
									% leng]];
						}
					}

					boolean bExist = false;
					for (int p = 0; p < varPerms.size(); p++) {
						int[] tmpVarPerm = varPerms.get(p);
						int[] tmpActPerm = actPerms.get(p);

						if (isSamePerm(tmpVarPerm, newVarPerm)
								&& isSamePerm(tmpActPerm, newActPerm)) {
							bExist = true;
							break;
						}
					}
					if (!bExist) {
						varPerms.add(newVarPerm);
						actPerms.add(newActPerm);
						bChange = true;
					}
				}
			}

		}

		printPerms(varPerms, actPerms);
		savePerms(varPerms, actPerms, filename);
	}

	public static void printPerms(ArrayList<int[]> varPerms,
			ArrayList<int[]> actPerms) {
		System.out.println(varPerms.size());
		for (int perm = 0; perm < varPerms.size(); perm++) {
			int[] varPerm = varPerms.get(perm);
			for (int i = 0; i < varPerm.length; i++) {
				System.out.print(varPerm[i] + " ");
			}
			int[] actPerm = actPerms.get(perm);
			for (int i = 0; i < actPerm.length; i++) {
				System.out.print(actPerm[i] + " ");
			}
			System.out.println();
		}
	}

	public static void savePerms(ArrayList<int[]> varPerms,
			ArrayList<int[]> actPerms, String filename) throws Exception {
		BufferedWriter bw = new BufferedWriter(new FileWriter(filename));
		bw.write(varPerms.size() + "\n");
		for (int perm = 0; perm < varPerms.size(); perm++) {
			int[] varPerm = varPerms.get(perm);
			for (int i = 0; i < varPerm.length; i++) {
				bw.write(varPerm[i] + " ");
			}
			int[] actPerm = actPerms.get(perm);
			for (int i = 0; i < actPerm.length; i++) {
				bw.write(actPerm[i] + " ");
			}
			bw.write("\n");
		}
		bw.close();
	}

	public static boolean isSamePerm(int[] perm1, int[] perm2) {
		if (perm1.length != perm2.length)
			return false;
		for (int i = 0; i < perm1.length; i++) {
			if (perm1[i] != perm2[i])
				return false;
		}
		return true;
	}

}
