package kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011;

import java.util.ArrayList;

import kr.ac.kaist.pomdp.symbolic.perseus.DD;
import kr.ac.kaist.pomdp.symbolic.perseus.DDleaf;
import kr.ac.kaist.pomdp.symbolic.perseus.HashMap;
import kr.ac.kaist.pomdp.symbolic.perseus.OP;
import kr.ac.kaist.pomdp.symbolic.perseus.SymbolicProblem;

/**
 * Search dd
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class Search {
	public static EdgeDd getEdgeDd(int horizon, SymbolicProblem p,
			SymbolicHsviIPPC2011 problem, DD ospProbs, DD oProbs) {
		int[][] config = new int[2][p.obsVarsPrimeId.length];
		HashMap hashtable = new HashMap(); // mapping from next belief to
											// EdgeDdLeaf

		hashtable.put(Edge.zeroEdge, EdgeDd.zero); // impossible next belief is
													// represented as
													// EdgeDd.zero
		return _getEdgeDd(horizon, p, problem, ospProbs, oProbs, 0, config,
				hashtable);
	}

	public static EdgeDd _getEdgeDd(int horizon, SymbolicProblem p,
			SymbolicHsviIPPC2011 problem, DD ospProbs, DD oProbs, int size,
			int[][] config, HashMap hashtable) {
		int ospHead = ospProbs.getVar();

		// ospProbs reached terminal or state var
		if (ospHead == 0 || ospHead < p.obsVarsPrimeId[0]) {
			// use config to determine observation probability
			// at this stage, oProb should have reached terminal
			double oprob = OP.eval(oProbs, config);

			if (oprob == 0) {
				return EdgeDd.zero;
			}

			// config
			int[] myconfig = new int[p.nObsVars];
			for (int i = 0; i < config[0].length && config[0][i] != 0; i++)
				myconfig[config[0][i] - p.obsVarsPrimeId[0]] = config[1][i];

			// next belief
			DD nextBelief = OP.primeVars(ospProbs, -p.nVars);
			nextBelief = OP.div(nextBelief, DDleaf.myNew(oprob));

			// make edge
			Edge e = new Edge();
			e.obsConfig = myconfig;
			e.mult = Common.getMult(e.obsConfig, p.obsVarsArity);
			e.obsProb = oprob;
			ArrayList nodePerm = problem.bringNode(horizon - 1, nextBelief);
			e.nextNode = (Node) nodePerm.get(0);
			e.invPerm = (int[][]) nodePerm.get(1);

			// System.out.println("obs:");
			// for (int i=0; i<e.obsConfig.length; i++)
			// System.out.print(e.obsConfig[i] + " ");
			// System.out.println();
			// System.out.println();
			//
			// for (int i=0; i<e.nextNode.b.length; i++)
			// e.nextNode.b[i].display();

			EdgeDd storedResult = (EdgeDd) hashtable.get(e);
			if (storedResult == null) {
				storedResult = EdgeDdLeaf.myNew(e);
				hashtable.put(e, storedResult);
			}

			return storedResult;
		} else {
			config[0][size] = ospHead;

			// recursion
			DD[] children = ospProbs.getChildren();
			EdgeDd[] edgeDdChildren = new EdgeDd[children.length];
			for (int i = 0; i < children.length; i++) {
				config[1][size] = i + 1;
				edgeDdChildren[i] = _getEdgeDd(horizon, p, problem,
						children[i], oProbs, size + 1, config, hashtable);
			}

			return EdgeDdNode.myNew(ospHead, edgeDdChildren);
		}
	}
}
