package kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011;

import java.io.Serializable;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;
import kr.ac.kaist.pomdp.symbolic.perseus.SymbolicProblem;
import kr.ac.kaist.pomdp.symbolic.perseus.DD;

/**
 * SymbolicSymmetry Class for making the graph to solve graph automorphism
 * problem
 * 
 * usage 1. SymbolicSymmetry create graph file of a problem ex)
 * "game_of_life_inst_pomdp__1.sperseus.pomdpGraph" 2. bliss find graph
 * automorphism using bliss software ex)
 * "./bliss game_of_life_inst_pomdp__1.sperseus.pomdpGraph" 3. PermGenerator
 * From the result of bliss, input the generators of bliss, and find all
 * permutations
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class SymbolicSymmetry implements Serializable {
	private static final long serialVersionUID = 8603599358545636973L;

	public static int nNodes;
	public static int nEdges;
	public static int nColors;

	public static int offset;

	public static void main(String[] args) throws Exception {

		String symFilename = "./data/final_comp/spudd_sperseus/game_of_life_inst_pomdp__1.sperseus";

		boolean printConsole = false;

		System.out.println("Symbolic filename : " + symFilename);
		SymbolicProblem p = new SymbolicProblem(symFilename, false);
		// p.display();

		// pomdp automorphism
		System.out.println(" *** POMDP Automorphism");
		init();
		if (printConsole) {
			makePomdpGraph(p, System.out, System.out);
		} else {
			String tmpFilenameNodes = symFilename + ".pomdpGraph.tmpNodes";
			PrintStream psNodes = new PrintStream(new FileOutputStream(
					tmpFilenameNodes));
			String tmpFilenameEdges = symFilename + ".pomdpGraph.tmpEdges";
			PrintStream psEdges = new PrintStream(new FileOutputStream(
					tmpFilenameEdges));

			makePomdpGraph(p, psNodes, psEdges);

			psNodes.close();
			psEdges.close();

			BufferedReader brNodes = new BufferedReader(new FileReader(
					tmpFilenameNodes));
			BufferedReader brEdges = new BufferedReader(new FileReader(
					tmpFilenameEdges));
			String outFilename = symFilename + ".pomdpGraph";
			PrintStream ps = new PrintStream(new FileOutputStream(outFilename));
			printMergePs(brNodes, brEdges, ps, symFilename, p);
			ps.close();
			brNodes.close();
			brEdges.close();

			File tmpFileNodes = new File(tmpFilenameNodes);
			tmpFileNodes.delete();
			File tmpFileEdges = new File(tmpFilenameEdges);
			tmpFileEdges.delete();
		}
		System.out.println();
	}

	public static void init() {
		nNodes = 1;
		nEdges = 1;
		nColors = 1;
		offset = 0;
	}

	public static void printMergePs(BufferedReader brNodes,
			BufferedReader brEdges, PrintStream ps, String symFilename,
			SymbolicProblem p) throws Exception {
		System.out.print("c " + symFilename + "\n");
		if (p != null)
			System.out.print("v " + "svar " + p.nStateVars + " ovar "
					+ p.nObsVars + " act " + p.nActions + "\n");
		System.out.print("p edge " + (nNodes - 1) + " " + (nEdges - 1) + "\n");
		System.out.print(" ** Writing (PO)MDP Graph for bliss...");

		ps.print("c " + symFilename + "\n");
		ps.print("p edge " + (nNodes - 1) + " " + (nEdges - 1) + "\n");

		String line = null;
		while ((line = brNodes.readLine()) != null) {
			ps.print(line + "\n");
		}
		while ((line = brEdges.readLine()) != null) {
			ps.print(line + "\n");
		}
		System.out.println(" complete");
	}

	public static void printVarNodesEdges(SymbolicProblem p,
			PrintStream psNodes, PrintStream psEdges, boolean bPomdp) {
		// state vars (1 ~ nStateVars: stateVars, nStateVars+1 ~ 2*nStateVars:
		// primed stateVars)
		for (int stVar = 0; stVar < p.nStateVars; stVar++) {
			psNodes.print("n " + nNodes + " " + nColors + "\n");
			nNodes++;
		}
		nColors++;
		for (int stVar = 0; stVar < p.nStateVars; stVar++) {
			psNodes.print("n " + nNodes + " " + nColors + "\n");
			nNodes++;
		}
		nColors++;
		// obs vars (2*nStateVars+1 ~ 2*nStateVars+nObsVars: primed obsVars)
		for (int obsVar = 0; obsVar < p.nObsVars; obsVar++) {
			psNodes.print("n " + nNodes + " " + nColors + "\n");
			if (!bPomdp)
				nColors++;
			nNodes++;
		}
		nColors++;
		// actions (2*nStateVars+nObsVars+1 ~ 2*nStateVars+nObsVars+nActions)
		for (int ac = 0; ac < p.nActions; ac++) {
			psNodes.print("n " + nNodes + " " + nColors + "\n");
			nNodes++;
		}
		nColors++;
		offset = nNodes - 1;

		nColors = 5;
		// state vars (1 ~ 2*nStateVars) (true=1, false = 2)
		for (int stVar = 0; stVar < 2 * p.nStateVars; stVar++) {
			psNodes.print("n " + nNodes + " " + nColors + "\n");
			psEdges.print("e " + nNodes + " " + (stVar / 2 + 1) + "\n");
			nNodes++;
			nEdges++;
		}
		nColors++;
		// primed state vars (2*nStateVars+1 ~ 4*nStateVars) (true=1, false = 2)
		for (int stVar = 0; stVar < 2 * p.nStateVars; stVar++) {
			psNodes.print("n " + nNodes + " " + nColors + "\n");
			psEdges.print("e " + nNodes + " " + (stVar / 2 + 1 + p.nStateVars)
					+ "\n");
			nNodes++;
			nEdges++;
		}
		nColors++;
		// obs vars (4*nStateVars+nActions+1 ~ 4*nStateVars+nActions +
		// 2*nObsVars) (true=1, false = 2)
		for (int obsVar = 0; obsVar < 2 * p.nObsVars; obsVar++) {
			psNodes.print("n " + nNodes + " " + nColors + "\n");
			if (!bPomdp)
				nColors++;
			psEdges.print("e " + nNodes + " "
					+ (obsVar / 2 + 1 + 2 * p.nStateVars) + "\n");
			nNodes++;
			nEdges++;
		}
		nColors++;
	}

	public static void printVarNodes(SymbolicProblem p, PrintStream psNodes) {
		// state vars (1 ~ 2*nStateVars) (true=1, false = 2)
		for (int stVar = 0; stVar < 2 * p.nStateVars; stVar++) {
			psNodes.print("n " + nNodes + " " + 1 + "\n");
			nNodes++;
		}
		// primed state vars (2*nStateVars+1 ~ 4*nStateVars) (true=1, false = 2)
		for (int stVar = 0; stVar < 2 * p.nStateVars; stVar++) {
			psNodes.print("n " + nNodes + " " + 2 + "\n");
			nNodes++;
		}
		// actions (4*nStateVars+1 ~ 4*nStateVars+nActions)
		for (int ac = 0; ac < p.nActions; ac++) {
			psNodes.print("n " + nNodes + " " + 3 + "\n");
			nNodes++;
		}
		nColors = 5;
	}

	public static void printObsVarNodes(SymbolicProblem p, PrintStream psNodes) {
		// obs vars (4*nStateVars+nActions+1 ~ 4*nStateVars+nActions +
		// 2*nObsVars) (true=1, false = 2)
		for (int obsVar = 0; obsVar < 2 * p.nObsVars; obsVar++) {
			psNodes.print("n " + nNodes + " " + 4 + "\n");
			nNodes++;
		}
		nColors = 5;
	}

	public static void makePomdpGraph(SymbolicProblem p, PrintStream psNodes,
			PrintStream psEdges) {

		System.out.println(" ** Make POMDP Graph (2)");

		// ex) key: "T%f8.6"+prob, value: color
		HashMap hm = new HashMap();

		printVarNodesEdges(p, psNodes, psEdges, true);

		for (int ac = 0; ac < p.nActions; ac++) {
			System.out.println("action = " + ac + " / " + p.nActions);
			ArrayList defaultList = new ArrayList();
			defaultList.add(1 + ac + p.nVars + p.nStateVars);

			DD[] transFn = p.actions[ac].transFn;
			DD rewFn = p.actions[ac].rewFn;
			DD[] obsFn = p.actions[ac].obsFn;

			// transition dds
			for (int i = 0; i < transFn.length; i++) {
				DD dd = transFn[i];
				int[] varSet = dd.getVarSet();
				int nVarSet = varSet.length;
				addAllPath(hm, dd, (ArrayList) defaultList.clone(), p, "T",
						psNodes, psEdges, varSet, nVarSet - 1);
			}

			// reward dd
			{
				int[] varSet = rewFn.getVarSet();
				int nVarSet = varSet.length;
				addAllPath(hm, rewFn, (ArrayList) defaultList.clone(), p, "R",
						psNodes, psEdges, varSet, nVarSet - 1);
			}

			// observation dds
			for (int i = 0; i < obsFn.length; i++) {
				DD dd = obsFn[i];
				int[] varSet = dd.getVarSet();
				int nVarSet = varSet.length;
				addAllPath(hm, dd, (ArrayList) defaultList.clone(), p, "Z",
						psNodes, psEdges, varSet, nVarSet - 1);
			}
		}

		printHashMap(hm);
	}

	public static void addAllPath(HashMap hm, DD dd, ArrayList list,
			SymbolicProblem p, String fName, PrintStream psNodes,
			PrintStream psEdges, int[] varSet, int varIndex) {

		DD[] children = null;
		int var = dd.getVar();
		if (varIndex >= 0) {
			if (var == varSet[varIndex]) {
				children = dd.getChildren();
			} else {
				var = varSet[varIndex];
				children = new DD[2];
				children[0] = dd;
				children[1] = dd;
			}
		}

		if (children != null) {
			// non-leaf node
			// only for boolean variables
			for (int i = 0; i < 2 /* children.length */; i++) {
				int tmpVar = var;
				// state vars
				if (tmpVar <= p.nStateVars) {
					tmpVar = (tmpVar - 1) * 2 + 1 + i;
				} else
				// state prime vars
				if (tmpVar > p.nVars && tmpVar <= p.nVars + p.nStateVars) {
					tmpVar -= p.nObsVars;
					tmpVar = (tmpVar - 1) * 2 + 1 + i;
				} else
				// obs (prime) vars
				if (tmpVar > p.nVars + p.nStateVars) {
					tmpVar -= (p.nVars + p.nStateVars);
					tmpVar = (tmpVar - 1) * 2 + 1 + i + p.nStateVars * 4;
				}
				tmpVar += offset;
				list.add(tmpVar);
				addAllPath(hm, children[i], list, p, fName, psNodes, psEdges,
						varSet, varIndex - 1);
				list.remove(list.size() - 1);
			}
		} else {
			// leaf node
			String key = makeStringKey(fName, dd.getVal());
			if (hm.get(key) != null) {
				int color = (Integer) hm.get(key);
				psNodes.print("n " + nNodes + " " + color + "\n");
				for (Object obj : list) {
					int tmpVar = (Integer) obj;
					psEdges.print("e " + nNodes + " " + tmpVar + "\n");
					nEdges++;
				}
				nNodes++;
			} else {
				hm.put(key, nColors);
				psNodes.print("n " + nNodes + " " + nColors + "\n");
				for (Object obj : list) {
					int tmpVar = (Integer) obj;
					psEdges.print("e " + nNodes + " " + tmpVar + "\n");
					nEdges++;
				}
				nNodes++;
				nColors++;
			}
		}
	}

	public static String makeStringKey(String fName, double key) {
		return String.format("%s%12.6f", fName, key);
	}

	public static void printHashMap(HashMap hm) {
		Iterator iterHm = hm.keySet().iterator();
		while (iterHm.hasNext()) {
			String key = (String) iterHm.next();
			System.out.println("key = " + key + ", \tcolor = " + hm.get(key));
		}
	}

}
