package kr.ac.kaist.pomdp.symbolic.perseus;

import java.io.PrintStream;
import java.lang.ref.WeakReference;

/**
 * DDleaf class - leaf node in ADD
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class DDleaf extends DD
{
	private static final long serialVersionUID = -7359878050328275041L;

	private double val; // value
	private int[][] config; // configuration

	/**
	 * DDleaf constructor
	 * 
	 * @param val
	 * 			input value
	 */
	private DDleaf(double val)
	{
		this.val = val;
		this.var = 0;
		this.config = null;
	}

	/**
	 * DDleaf constructor
	 * 
	 * @param val
	 * 			input value
	 * @param config
	 * 			input configuration
	 */
	private DDleaf(double val, int[][] config)
	{
		this.val = val;
		this.var = 0;
		this.config = config;
	}

	/**
	 * get new DDleaf instance using leafHashtable
	 * 
	 * @param val
	 * 			input value
	 * @return new DDleaf instance
	 */
	public static DD myNew(double val)
	{
		if (val == 0 && DD.zero != null)
			return DD.zero;

		if (val == 1 && DD.one != null)
			return DD.one;

		// create new leaf
		DDleaf leaf = new DDleaf(val);

		// try to lookup leaf in leafHashtable 
		WeakReference storedLeaf = (WeakReference) Global.leafHashtable.get(leaf);

		if (storedLeaf != null)
			return (DDleaf) storedLeaf.get();

		// store leaf in leafHashtable
		Global.leafHashtable.put(leaf, new WeakReference<DD>(leaf));
		return leaf;
	}

	/**
	 * get new DDleaf instance using leafHashtable
	 * 
	 * @param val
	 * 			input value
	 * @param config
	 * 			input configuration
	 * @return new DDleaf instance
	 */
	public static DD myNew(double val, int[][] config)
	{
		if (val == 0 && config == null && DD.zero != null)
			return DD.zero;
		
		if (val == 1 && config == null && DD.one != null)
			return DD.one;

		// create new leaf
		DDleaf leaf = new DDleaf(val, config);

		// try to lookup leaf in leafHashtable
		WeakReference storedLeaf = (WeakReference) Global.leafHashtable.get(leaf);

		if (storedLeaf != null)
			return (DDleaf) storedLeaf.get();

		// store leaf in leafHashtable
		Global.leafHashtable.put(leaf, new WeakReference<DD>(leaf));
		return leaf;
	}

	/**
	 * get variable set
	 * there is no information about variables in a leaf node.
	 */
	public int[] getVarSet()
	{
		return new int[0];
	}

	/**
	 * get value
	 */
	public double getVal()
	{
		return val;
	}

	/**
	 * get configuration
	 */
	public int[][] getConfig()
	{
		return config;
	}

	/**
	 * get the number of leaves of this node
	 * always return 1 
	 */
	public int getNumLeaves()
	{
		return 1;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj)
	{

		if (obj.getClass() != getClass())
			return false;
		DDleaf leaf = (DDleaf) obj;

		if (val == leaf.val && Config.equals(config, leaf.config))
			return true;
		else
			return false;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode()
	{
		Double valD = new Double(val);
		return valD.hashCode() + Config.hashCode(config);
	}

	/* (non-Javadoc)
	 * @see kr.ac.kaist.pomdp.symbolic.DD#store()
	 */
	public DD store()
	{
		return DDleaf.myNew(val, config);
	}

	/* (non-Javadoc)
	 * @see kr.ac.kaist.pomdp.symbolic.DD#display(java.lang.String)
	 */
	public void display(String space)
	{
		System.out.println(space + "leaf: " + Double.toString(val) + "  " + Config.toString(config)
				+ " (id:" + this + ")");
	}

	/* (non-Javadoc)
	 * @see kr.ac.kaist.pomdp.symbolic.DD#display(java.lang.String, java.lang.String)
	 */
	public void display(String space, String prefix)
	{
		System.out.println(space + prefix + Double.toString(val) + "  " + Config.toString(config)
				+ " (id:" + this + ")");
	}

	/* (non-Javadoc)
	 * @see kr.ac.kaist.pomdp.symbolic.DD#printSpuddDD(java.io.PrintStream)
	 */
	public void printSpuddDD(PrintStream ps)
	{
		ps.print("(" + Double.toString(val) + ")");
	}
}
