package kr.ac.kaist.pomdp.symbolic.perseus;

import java.lang.ref.WeakReference;
import java.util.Iterator;
import java.util.Random;
import java.util.Set;

/**
 * Global class
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class Global
{
	public static int[] varDomSize = null;

	public static String[] varNames = null;
	public static String[][] valNames = null;

	// hash tables
	public static CacheMap leafHashtable = new CacheMap();
	public static CacheMap nodeHashtable = new CacheMap();
	public static CacheMap addHashtable = new CacheMap();
	public static CacheMap multHashtable = new CacheMap();
	public static CacheMap maxHashtable = new CacheMap();
	public static CacheMap minHashtable = new CacheMap();
	public static CacheMap dotProductHashtable = new CacheMap();
	public static CacheMap nEdgesHashtable = new CacheMap();
	public static CacheMap nLeavesHashtable = new CacheMap();
	public static CacheMap nNodesHashtable = new CacheMap();

	// random number generator
	public static Random random = new Random();

	public static void setVarDomSize(int[] newVarDomSize)
	{
		Global.varDomSize = newVarDomSize;
	}

	public static void setVarNames(String[] newVarNames)
	{
		Global.varNames = newVarNames;
	}

	public static void setValNames(int varId, String[] newValNames)
	{
		if (Global.valNames == null)
		{
			Global.valNames = new String[varId][];
			Global.valNames[varId - 1] = newValNames;
		}
		else if (Global.valNames.length < varId)
		{
			String[][] tempValNames = new String[varId][];
			for (int i = 0; i < Global.valNames.length; i++)
			{
				tempValNames[i] = Global.valNames[i];
			}
			tempValNames[varId - 1] = newValNames;
			Global.valNames = tempValNames;
		}
		else
		{
			Global.valNames[varId - 1] = newValNames;
		}
	}

	public static void clearHashtables()
	{
		Global.leafHashtable.clear();
		Global.nodeHashtable.clear();
		Global.addHashtable.clear();
		Global.multHashtable.clear();
		Global.maxHashtable.clear();
		Global.minHashtable.clear();
		Global.dotProductHashtable.clear();
		Global.nEdgesHashtable.clear();
		Global.nLeavesHashtable.clear();
		Global.nNodesHashtable.clear();
		Global.leafHashtable.put(DD.zero, new WeakReference<DD>(DD.zero));
		Global.leafHashtable.put(DD.one, new WeakReference<DD>(DD.one));
	}

	public static void newHashtables()
	{
		Global.leafHashtable = new CacheMap();
		Global.nodeHashtable = new CacheMap();
		Global.addHashtable = new CacheMap();
		Global.multHashtable = new CacheMap();
		Global.maxHashtable = new CacheMap();
		Global.minHashtable = new CacheMap();
		Global.dotProductHashtable = new CacheMap();
		Global.nEdgesHashtable = new CacheMap();
		Global.nLeavesHashtable = new CacheMap();
		Global.nNodesHashtable = new CacheMap();
		Global.leafHashtable.put(DD.zero, new WeakReference<DD>(DD.zero));
		Global.leafHashtable.put(DD.one, new WeakReference<DD>(DD.one));
	}

	public static int[] getKeyHashCodeSet(HashMap hashMap)
	{
		Set keySet = hashMap.keySet();
		Iterator iterator = keySet.iterator();
		int[] hashCodeCollection = new int[hashMap.size()];
		int i = 0;
		while (iterator.hasNext())
		{
			hashCodeCollection[i] = iterator.next().hashCode();
			i += 1;
		}
		return hashCodeCollection;
	}

	public static void main(String args[])
	{

		int N = 100;
		int iter = 1;

		Random numberGenerator = new Random(10101);

		for (long j = 0; j < iter; j++)
		{

			DD dd1 = DDleaf.myNew(numberGenerator.nextDouble());
			for (int i = 1; i <= N; i++)
			{
				DD[] children = new DD[3];
				if (numberGenerator.nextInt(2) == 0)
				{
					children[0] = DDleaf.myNew(numberGenerator.nextDouble());
					children[1] = dd1;
					children[2] = dd1;
				}
				else
				{
					children[0] = dd1;
					children[1] = dd1;
					children[2] = DDleaf.myNew(numberGenerator.nextDouble());
				}
				dd1 = DDnode.myNew(i, children);
			}

			DD dd2 = DDleaf.myNew(numberGenerator.nextDouble());
			for (int i = 1; i <= N; i++)
			{
				DD[] children = new DD[3];
				if (numberGenerator.nextInt(2) == 0)
				{
					children[0] = DDleaf.myNew(numberGenerator.nextDouble());
					children[1] = dd2;
					children[2] = dd2;
				}
				else
				{
					children[0] = dd2;
					children[1] = dd2;
					children[2] = DDleaf.myNew(numberGenerator.nextDouble());
				}
				dd2 = DDnode.myNew(i, children);
			}

			DD[] ddArray = new DD[2];
			ddArray[0] = dd1;
			ddArray[1] = dd2;

			int[] varSet = new int[N];
			int[][] config = new int[2][N];
			for (int i = 1; i <= N; i++)
			{
				varSet[i - 1] = i;
				config[0][i - 1] = i;
				config[1][i - 1] = i % 2 + 1;
			}

			//System.out.println("dd1.display");
			//dd1.display();
			Global.multHashtable.clear();
			Global.addHashtable.clear();
			Global.leafHashtable.clear();
			Global.nodeHashtable.clear();
		}
		Global.dotProductHashtable.clear();
		Global.multHashtable.clear();
		Global.addHashtable.clear();
		Global.leafHashtable.clear();
		Global.nodeHashtable.clear();
		System.out.println("done");
	}
}
