package kr.ac.kaist.pomdp.symbolic.perseus;

import java.util.ArrayList;
import java.util.Random;

import kr.ac.kaist.pomdp.utils.Util;

/**
 * SymbolicBeliefPoint class
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class SymbolicBeliefPoints
{

	//double threshold;
	Random rand = new Random(0);
	DD[][] belRegion;
	SymbolicProblem p;

	public SymbolicBeliefPoints(DD initial, SymbolicProblem POMDP, int nBelStates, Random r)
	{

		rand = r;
		//threshold = 0.001;

		int[] stateVarsId = new int[POMDP.nStateVars];
		int[] stateVarsPrimeId = new int[POMDP.nStateVars];
		int[] obsVarsId = new int[POMDP.nObsVars];
		int[] obsVarsPrimeId = new int[POMDP.nObsVars];

		for (int varId = 0; varId < POMDP.nStateVars; varId++)
		{
			stateVarsId[varId] = POMDP.stateVars[varId].id;
			stateVarsPrimeId[varId] = POMDP.stateVars[varId].id + POMDP.nVars;
			//System.out.println(stateVarsId[varId]);
		}
		for (int varId = 0; varId < POMDP.nObsVars; varId++)
		{
			obsVarsId[varId] = POMDP.obsVars[varId].id;
			obsVarsPrimeId[varId] = POMDP.obsVars[varId].id + POMDP.nVars;
		}

		// factorize initBelState
		DD[] temp = new DD[POMDP.nStateVars];

		for (int varId = 0; varId < POMDP.nStateVars; varId++)
		{
			int[] vars = Util.elimOneVar(stateVarsId, varId);
			vars = MySet.remove(vars, 0);
			temp[varId] = OP.addMultVarElim(initial, vars);
		}
		DD[] initBelState = temp;

		int nSteps = 50;
		DD[] nextBelState = new DD[POMDP.nStateVars];
		belRegion = new DD[nBelStates][];
		int count = 0;
		belRegion[count] = initBelState;

		//----- test printing -----
		System.out.println("[" + (count + 1) + " th belief ]");
		//for(int j=0;j<belRegion[count].length;j++) {					
		//			belRegion[count][j].display();
		//}

		while (count < nBelStates - 1)
		{
			DD[] belState = new DD[initBelState.length];

			System.out.println("restart !");
			for (int i = 0; i < initBelState.length; i++)
			{
				belState[i] = initBelState[i].store();
			}

			for (int stepId = 0; stepId < nSteps; stepId++)
			{
				// sample action
				int actId = (int) (POMDP.nActions * rand.nextDouble());
				//if(actId>6) break;
				//stem.out.println("act: "+actId);

				// sample observation				
				DD[] dds;
				dds = OP.concat(belState, POMDP.actions[actId].transFn);
				dds = OP.concat(dds, POMDP.actions[actId].obsFn);

				DD obsDist = OP.addMultVarElim(dds, Util.concat(stateVarsId, stateVarsPrimeId));
				int[][] obsConfig = OP.sampleMultinomial(obsDist, obsVarsPrimeId);
				DD[] restrictedObsFn = OP.restrictN(POMDP.actions[actId].obsFn, obsConfig);

				// update belstate
				for (int varId = 0; varId < POMDP.nStateVars; varId++)
				{
					DD[] dds2;
					dds2 = OP.concat(belState, POMDP.actions[actId].transFn);
					dds2 = OP.concat(dds2, restrictedObsFn);

					// [stateVarsPrimeId([1:varId-1,varId+1:end]), stateVarsId]
					int[] vars = new int[POMDP.nStateVars * 2];
					int i_vars = 0;
					for (int i = 0; i < varId; i++)
					{
						vars[i_vars] = stateVarsPrimeId[i];
						i_vars++;
					}
					for (int i = varId + 1; i < POMDP.nStateVars; i++)
					{
						vars[i_vars] = stateVarsPrimeId[i];
						i_vars++;
					}
					for (int i = 0; i < POMDP.nStateVars; i++)
					{
						vars[i_vars] = stateVarsId[i];
						i_vars++;
					}
					vars = MySet.remove(vars, 0);

					nextBelState[varId] = OP.addMultVarElim(dds2, vars);

					double[] dvars = new double[1];
					dvars[0] = 0;
					nextBelState[varId] = OP.approximate(nextBelState[varId], 0.000001, dvars);

					vars = new int[1];
					vars[0] = stateVarsPrimeId[varId];
					nextBelState[varId] =
							OP.div(nextBelState[varId],
									OP.addMultVarElim(nextBelState[varId], vars));

				}
				belState = OP.primeVarsN(nextBelState, -POMDP.nVars);

				// compare belstate
				if (!containSameBelief(belRegion, count + 1, belState, 0.3))
				{ // add belstate to belregion
					count++;
					belRegion[count] = belState;
					System.out.println("[" + (count) + " th belief ] action "
							+ POMDP.actions[actId].name);
					printBelief(belState);
				}

				// ---- test printing -----
				//if((count+1)%10==0) {

				/*for(int j=0;j<belRegion[count].length;j++) {					
					belRegion[count][j].display();
				}*/
				//}

				//if ((count+1)%10==0) System.out.println((count+1)+" belief states sampled");
				if (count >= nBelStates - 1)
					return;

				Global.newHashtables();

			}
		}

	}

	public SymbolicBeliefPoints(DD initial, SymbolicProblem POMDP, int nBelStates, double epsilon)
	{

		//threshold = 0.001;
		p = POMDP;
		Global.random.setSeed(1);

		int[] stateVarsId = new int[POMDP.nStateVars];
		int[] stateVarsPrimeId = new int[POMDP.nStateVars];
		int[] obsVarsId = new int[POMDP.nObsVars];
		int[] obsVarsPrimeId = new int[POMDP.nObsVars];

		for (int varId = 0; varId < POMDP.nStateVars; varId++)
		{
			stateVarsId[varId] = POMDP.stateVars[varId].id;
			stateVarsPrimeId[varId] = POMDP.stateVars[varId].id + POMDP.nVars;
			//System.out.println(stateVarsId[varId]);
		}
		for (int varId = 0; varId < POMDP.nObsVars; varId++)
		{
			obsVarsId[varId] = POMDP.obsVars[varId].id;
			obsVarsPrimeId[varId] = POMDP.obsVars[varId].id + POMDP.nVars;
		}

		// factorize initBelState
		DD[] temp = new DD[POMDP.nStateVars];

		for (int varId = 0; varId < POMDP.nStateVars; varId++)
		{
			int[] vars = Util.elimOneVar(stateVarsId, varId);
			vars = MySet.remove(vars, 0);
			temp[varId] = OP.addMultVarElim(initial, vars);
		}
		DD[] initBelState = temp;
		DD[] nextBelState = new DD[POMDP.nStateVars];
		belRegion = new DD[nBelStates][];

		belRegion[0] = initBelState;

		//----- test printing -----

		DD[][] belState = new DD[POMDP.nActions][initBelState.length];
		int currentSize = 1;
		int lastRestart = 0;

		while (currentSize < nBelStates && lastRestart < 100)
		{

			int count = currentSize;

			for (int beliefId = 0; beliefId < count; beliefId++)
			{
				for (int actId = 0; actId < POMDP.nActions; actId++)
				{

					belState[actId] = belRegion[beliefId];

					// sample observation				
					DD[] dds;
					dds = OP.concat(belState[actId], POMDP.actions[actId].transFn);
					dds = OP.concat(dds, POMDP.actions[actId].obsFn);

					DD obsDist = OP.addMultVarElim(dds, Util.concat(stateVarsId, stateVarsPrimeId));
					int[][] obsConfig = OP.sampleMultinomial(obsDist, obsVarsPrimeId);
					DD[] restrictedObsFn = OP.restrictN(POMDP.actions[actId].obsFn, obsConfig);

					// update belstate
					for (int varId = 0; varId < POMDP.nStateVars; varId++)
					{
						DD[] dds2;
						dds2 = OP.concat(belState[actId], POMDP.actions[actId].transFn);
						dds2 = OP.concat(dds2, restrictedObsFn);

						// [stateVarsPrimeId([1:varId-1,varId+1:end]), stateVarsId]
						int[] vars = new int[POMDP.nStateVars * 2];
						int i_vars = 0;
						for (int i = 0; i < varId; i++)
						{
							vars[i_vars] = stateVarsPrimeId[i];
							i_vars++;
						}
						for (int i = varId + 1; i < POMDP.nStateVars; i++)
						{
							vars[i_vars] = stateVarsPrimeId[i];
							i_vars++;
						}
						for (int i = 0; i < POMDP.nStateVars; i++)
						{
							vars[i_vars] = stateVarsId[i];
							i_vars++;
						}
						vars = MySet.remove(vars, 0);

						nextBelState[varId] = OP.addMultVarElim(dds2, vars);

						double[] dvars = new double[1];
						dvars[0] = 0;
						nextBelState[varId] = OP.approximate(nextBelState[varId], 0.000001, dvars);

						vars = new int[1];
						vars[0] = stateVarsPrimeId[varId];
						nextBelState[varId] =
								OP.div(nextBelState[varId],
										OP.addMultVarElim(nextBelState[varId], vars));
						//nextBelState[varId].display();

					}
					belState[actId] = OP.primeVarsN(nextBelState, -POMDP.nVars);
					//System.out.println("----------------------------");
				}
				// TODO: pick the farthest belstate
				ArrayList result = farthestBelief(belRegion, currentSize, belState);

				// add belstate to belregion
				int bestActId = (Integer) result.get(0);
				double distance = (Double) result.get(1);

				lastRestart++;

				if (distance > epsilon)
				{
					belRegion[currentSize] = belState[bestActId];
					//System.out.println("act: "+p.actions[bestActId].name+", sum of max norm distance: "+distance);
					//printBelief(belState[bestActId]);					

					currentSize++;
					if (currentSize % 100 == 0)
						System.out.println(currentSize + " points sampled");
					lastRestart = 0;
				}

				if (currentSize >= nBelStates)
					return;

				Global.newHashtables();
			}

		}
		DD[][] newBeliefs = new DD[currentSize][];
		for (int i = 0; i < currentSize; i++)
		{
			newBeliefs[i] = belRegion[i];
			//newBeliefs[i][0].display();
		}

	}

	public ArrayList farthestBelief(DD[][] sampledPoints, int count, DD[][] belState)
	{
		ArrayList result = new ArrayList();

		double dist[] = new double[p.nActions];

		for (int a = 0; a < p.nActions; a++)
		{
			double distToB = Double.POSITIVE_INFINITY;
			for (int beliefId = 0; beliefId < count; beliefId++)
			{
				DD[] current = sampledPoints[beliefId];
				double distance = DDdistance(current, belState[a]);
				distToB = Math.min(distToB, distance);
			}
			dist[a] = distToB;
		}
		result.add(Util.argmax(dist));
		result.add(Util.max(dist));
		return result;
	}

	public double DDdistance(DD[] current, DD[] target)
	{
		double sum = 0;
		for (int i = 0; i < current.length; i++)
		{
			sum += OP.maxAll(OP.abs(OP.sub(current[i], target[i])));
		}
		return sum;
	}

	public void printBelief(DD[] belief)
	{
		for (int i = 0; i < belief.length; i++)
		{
			belief[i].display();
		}
	}

	public boolean containSameBelief(DD[][] beliefSet, int count, DD[] belState, double ep)
	{
		for (int i = 0; i < count; i++)
		{
			if (DDdistance(beliefSet[i], belState) < ep)
				return true;
		}
		return false;
	}
}
