package kr.ac.kaist.pomdp.symbolic.perseus;

import kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011.Common;
import kr.ac.kaist.pomdp.utils.*;
import java.util.*;

public class SymbolicPerseus
{

	private SymbolicProblem p;
	private SymbolicBeliefPoints sbp;
	public DD[][] belRegionArray;
	private int beliefSize;
	private int nIterations = 30;
	private Random rand = new Random(0);
	public Vector<Alpha> alphas;
	public int randLimit;
	public double bellmanErr;
	int maxSize;

	public double[] dzero()
	{
		double[] result = new double[1];
		result[0] = 0.0;
		return result;
	}

	public SymbolicPerseus(SymbolicProblem p, SymbolicBeliefPoints sbp, int nIter, Random r,
			int randLimit)
	{
		this.p = p;
		this.sbp = sbp;
		beliefSize = sbp.belRegion.length;
		nIterations = nIter;
		this.randLimit = randLimit;
		rand = r;
		bellmanErr = 20 * p.tolerance;
		alphas = new Vector();

		maxSize = 1000;
		belRegionArray = sbp.belRegion;
	}

	public void run()
	{
		System.out.println("running symbolic Perseus\n");
		dd_boundedPerseus(nIterations);
		System.out.println("iteration ended");
	}

	public void dd_boundedPerseus(int maxIterations)
	{

		//Vector<DD> alphaVectors = new Vector(maxSize);
		//Vector<Integer> newActions = null;//		
		//Vector<double[]> currentPointBasedValues = new Vector(maxSize);//[sbp.belRegion.length];
		//Vector<double[]> newPointBasedValues = null;// = new Vector(maxSize);//[sbp.belRegion.length];
		//Vector<DD> newAlphaVectors = null;// = new Vector(maxSize);

		// MODIFIED 11.09
		Vector<Alpha> alphaVectors = new Vector(maxSize);
		double[] currentPointBasedValues = new double[beliefSize];
		//double[] newPointBasedValues = new double[sbp.belRegion.length];                   ;
		//Vector<Alpha> newAlphaVectors = null;
		//Vector remove_candidate = new Vector(sbp.belRegion.length);

		//double[] diff;
		//double[] dotProducts;

		// MODIFIED 11.09
		/*double minRew = Double.POSITIVE_INFINITY;
		for(int a=0; a<p.nActions; a++) {
			minRew = Math.min(minRew, OP.minAll(p.actions[a].rewFn));
		}*/
		Alpha initAlpha = new Alpha(0, DD.zero);
		alphaVectors.add(initAlpha);

		// compute current point-based values		
		currentPointBasedValues =
				OP.factoredExpectationSparseNoMem(belRegionArray, alphaVectors.get(0).v);
		initAlpha.pbv = currentPointBasedValues;

		// start iteration		
		for (int stepId = 0; stepId < maxIterations; stepId++)
		{
			// precompute primedV
			DD[] alphaArray = Vector2Array(alphaVectors);
			DD[] primedV = new DD[alphaArray.length];

			System.out.println("stepId " + stepId);
			for (int i = 0; i < alphaArray.length; i++)
			{
				primedV[i] = OP.primeVars(alphaArray[i], p.nVars);
			}

			ArrayList result =
					epochIteration(alphaVectors, alphaArray, primedV, currentPointBasedValues);
			alphaVectors = (Vector) result.get(0);
			currentPointBasedValues = (double[]) result.get(1);
		}

		alphas = alphaVectors;
	}

	// MODIFIED 11.09
	public ArrayList epochIteration(Vector<Alpha> curAlphaVectors, DD[] curAlphaDDs,
			DD[] curPrimeDDs, double[] curPBV)
	{
		ArrayList result = new ArrayList();
		double[] newPBV = new double[beliefSize];
		for (int i = 0; i < beliefSize; i++)
			newPBV[i] = curPBV[i];
		Vector<Alpha> newAlphaVectors = new Vector();

		double tolerance = p.tolerance;

		// TODO: is this right?
		double maxAbsVal =
				Math.max(Util.maxAbs(OP.minAllN(curAlphaDDs)), Util.maxAbs(OP.maxAllN(curAlphaDDs)));
		maxAbsVal = Math.max(maxAbsVal, 1e-10);

		System.out.println("maxAbsVal = " + maxAbsVal);
		System.out.println("tolerance = " + tolerance);

		int count = 0;
		int nDpBackups = 0;
		Vector<Integer> stillLower = randperm(beliefSize);

		boolean pointStillLower = true;
		while (pointStillLower)
		{
			Global.newHashtables();
			count = count + 1;
			int choice;
			int target;

			// Randomize: # of backup of each iteration limited
			//if (nDpBackups >= randLimit*alphaVectors.size()) break;

			// Randomly pick a belief point for PB backup
			if (newAlphaVectors.size() == 0)
			{
				choice = 0;
				target = 0;
			}
			else
			{
				//stillLower.clear();
				/*for(int diffId=0;diffId<newPBV.length; diffId++) {
					if(newPBV[diffId]<curPBV[diffId]+tolerance) {
						stillLower.add(diffId);
					}
				}*/
				Iterator<Integer> it = stillLower.iterator();
				while (it.hasNext())
				{
					int pbId = it.next();
					if (newPBV[pbId] >= curPBV[pbId] + tolerance)
					{
						it.remove();
					}
				}

				if (stillLower.isEmpty())
				{
					pointStillLower = false;
					break;
				}
				choice = (int) (stillLower.size() * rand.nextDouble());
				target = (Integer) stillLower.get(choice);
			}
			stillLower.remove(choice);

			// DEBUG:
			/*Iterator<Integer> it = stillLower.iterator();
			while(it.hasNext()) {
				int pbId = it.next();
				System.out.print(pbId+" ");
			}
			System.out.println();*/

			// Point-based backup until all points improved
			if (newAlphaVectors.size() < 1 || newPBV[target] - curPBV[target] < tolerance)
			{

				// dpBackup
				ArrayList alist =
						dd_dpBackup(sbp.belRegion[target], curAlphaDDs, curPrimeDDs, maxAbsVal);

				DD newVector = (DD) alist.get(0);
				newVector = OP.approximate(newVector, bellmanErr / 10, dzero());
				int actId = (Integer) alist.get(2);
				Alpha newAlpha = new Alpha(actId, newVector);
				nDpBackups++;

				//newAlpha.v.display();

				// merge and trim
				double[] newValues = OP.factoredExpectationSparseNoMem(belRegionArray, newVector);
				newAlpha.pbv = newValues;
				newAlphaVectors.add(newAlpha);

				for (int bpId = 0; bpId < beliefSize; bpId++)
				{
					if (newPBV[bpId] < newValues[bpId])
					{
						newPBV[bpId] = newValues[bpId];
					}
				}
				//newPointBasedValues.add(newValues);
				//newActions.add(actId);

				// FOR DEBUG
				/*for(int tmpId=0; tmpId<beliefSize; tmpId++) {
					System.out.print(curPBV[tmpId]+" ");
				}
				System.out.println();
				for(int tmpId=0; tmpId<beliefSize; tmpId++) {
					System.out.print(newPBV[tmpId]+" ");
				}
				System.out.println();
				System.out.println("target: " + target);*/

				//double improvement = maxdiff(target, newPointBasedValues, currentPointBasedValues) 
				double improvement = newPBV[target] - curPBV[target];
				tolerance = Math.max(1e-10, Math.min(tolerance, improvement));

				//System.out.println("improvement: "+improvement);
				//System.out.println("tolerance: "+tolerance);

				// trim alpha vectors to improve too small amount
				ArrayList vlist = trim(newAlphaVectors, newPBV, tolerance);
				newAlphaVectors = (Vector) vlist.get(0);
				newPBV = (double[]) vlist.get(1);

				if (newAlphaVectors.size() >= maxSize)
					break;

			}

		}

		// statistics
		//double[] bi = difference(newPointBasedValues, currentPointBasedValues, 0, sbp.belRegion.length);
		//double bestImprovement = Util.max(bi);
		//double worstDecline = Util.min(bi);
		double bestImprovement = Double.NEGATIVE_INFINITY;
		double worstDecline = Double.POSITIVE_INFINITY;
		for (int bpId = 0; bpId < beliefSize; bpId++)
		{
			double tmp_val = newPBV[bpId] - curPBV[bpId];
			if (tmp_val > bestImprovement)
				bestImprovement = tmp_val;
			if (tmp_val < worstDecline)
				worstDecline = tmp_val;
		}

		double initValue = newPBV[0];
		double initImprovement = newPBV[0] - curPBV[0];
		System.out.println("\nbest Improvement = " + bestImprovement + "  worstDecline = "
				+ worstDecline + "  initImprovement = " + initImprovement + "\n" + "  initValue = "
				+ initValue);
		System.out.println("# of vectors = " + newAlphaVectors.size() + ", # of DP backups = "
				+ nDpBackups + "\n");

		result.add(newAlphaVectors);
		result.add(newPBV);
		bellmanErr = Math.min(10, Math.max(bestImprovement, -worstDecline));
		return result;
	}

	public ArrayList getAlphas()
	{
		ArrayList result = new ArrayList(2);
		java.util.Vector<DD> alphaVectors = new java.util.Vector();
		java.util.Vector<Integer> policy = new java.util.Vector();
		for (int i = 0; i < alphas.size(); i++)
		{
			alphaVectors.add(alphas.get(i).v);
			policy.add(alphas.get(i).a);
		}
		result.add(alphaVectors);
		result.add(policy);
		return result;
	}

	public double maxdiff(int target, Vector a, Vector b)
	{
		double max_a = Double.NEGATIVE_INFINITY;
		double max_b = Double.NEGATIVE_INFINITY;

		if (a == null)
			max_a = 0;
		else
		{
			for (int i = 0; i < a.size(); i++)
			{
				max_a = Math.max(max_a, ((double[]) a.get(i))[target]);
			}
		}
		if (b == null)
			max_b = 0;
		else
		{
			for (int i = 0; i < b.size(); i++)
			{
				max_b = Math.max(max_b, ((double[]) b.get(i))[target]);
			}
		}
		return max_a - max_b;
	}

	public double[] difference(Vector a, Vector b, double tolerance, int len)
	{
		double[] diff = new double[len];
		double[] max_a_column = new double[len];
		double[] max_b_column = new double[len];

		double[] a_column = new double[a.size()];
		double[] b_column = new double[a.size()];

		for (int i = 0; i < len; i++)
		{
			max_a_column[i] = Double.NEGATIVE_INFINITY;
			max_b_column[i] = Double.NEGATIVE_INFINITY;
		}

		// for a
		for (int i = 0; i < a.size(); i++)
		{
			a_column = (double[]) a.get(i);
			for (int j = 0; j < len; j++)
			{
				max_a_column[j] = Math.max(max_a_column[j], a_column[j]);
			}
		}
		// for b
		if (b == null)
			return a_column;
		for (int i = 0; i < b.size(); i++)
		{
			b_column = (double[]) b.get(i);
			for (int j = 0; j < len; j++)
			{
				max_b_column[j] = Math.max(max_b_column[j], b_column[j]);
			}
		}

		for (int i = 0; i < len; i++)
			diff[i] = max_a_column[i] - max_b_column[i] + tolerance;
		return diff;
	}

	// 0 ~ (length-1) ���ڷ� �̷����?���ڿ�; �߻�
	public Vector randperm(int length)
	{
		Vector result = new Vector(length);
		Vector seq = new Vector(length);
		for (int i = 0; i < length; i++)
		{
			seq.add(i);
		}
		for (int i = 0; i < length; i++)
		{
			int m = (int) (seq.size() * rand.nextDouble());
			result.add(seq.get(m));
			//result.add(seq.get(i));
			seq.remove(m);
		}
		return result;
	}

	//	 0 ~ (length-1) ���ڷ� �̷����?���ڿ�; �߻�
	public Vector<Integer> seqperm(int length)
	{
		Vector result = new Vector(length);
		for (int i = 0; i < length; i++)
		{
			result.add(i);
		}
		return result;
	}

	public ArrayList dd_dpBackup(DD[] belState, DD[] alphaVectors, DD[] primedV, double maxAbsVal)
	{
		ArrayList result = new ArrayList(3);

		//-------------------------------------------------------------
		// i) find best action and observation strategy at belState
		//-------------------------------------------------------------

		if (primedV.length == 1)
		{
			primedV = new DD[alphaVectors.length];
			for (int alphaId = 0; alphaId < alphaVectors.length; alphaId++)
			{
				primedV[alphaId] = OP.primeVars(alphaVectors[alphaId], p.nVars);
			}
		}

		// get next unnormailzed belStates
		ArrayList blist = dd_oneStepNZPrimeBelStates(belState, true, p.tolerance / maxAbsVal);

		Vector nextBelStates = (Vector) blist.get(0);
		Vector[] nzObsIds = (Vector[]) blist.get(1);

		// precompute obsVals
		double[][][] obsVals = new double[p.nActions][][];
		for (int actId = 0; actId < p.nActions; actId++)
		{
			if (!nzObsIds[actId].isEmpty())
			{
				obsVals[actId] =
						OP.factoredExpectationSparseNoMem((DD[][]) nextBelStates.get(actId),
								primedV);
			}
		}

		int[] obsStrat = new int[p.nObservations];
		int[] bestObsStrat = new int[p.nObservations];
		double bestValue = Double.NEGATIVE_INFINITY;
		int bestActId = 0;

		for (int actId = 0; actId < p.nActions; actId++)
		{
			// compute immediate reward
			double actValue = 0.0;
			actValue += OP.factoredExpectationSparseNoMem(belState, p.actions[actId].rewFn);

			// compute obsStrat
			for (int obsId = 0; obsId < p.nObservations; obsId++)
			{
				obsStrat[obsId] = 0;
			}
			double[] obsValues = new double[p.nObservations];

			for (int obsPtr = 0; obsPtr < nzObsIds[actId].size(); obsPtr++)
			{
				int obsId = (Integer) nzObsIds[actId].get(obsPtr);
				double obsProb = ((DD[][]) nextBelStates.get(actId))[obsPtr][p.nStateVars].getVal();

				double alphaValue = Util.max(obsVals[actId][obsPtr]);
				int obsStratObsId = Util.argmax(obsVals[actId][obsPtr]);

				obsStrat[obsId] = obsStratObsId;
				obsValues[obsId] = obsProb * alphaValue;
			}

			actValue += (Util.sum(obsValues) * p.discFact);

			// adjust bestValue
			if (actValue > bestValue)
			{
				bestValue = actValue;
				bestActId = actId;
				for (int obsId = 0; obsId < bestObsStrat.length; obsId++)
					bestObsStrat[obsId] = obsStrat[obsId];
			}
		}
		//-------------------------------------------------------------
		// ii) construct corresponding alpha vector
		//-------------------------------------------------------------
		DD newAlpha = DD.zero;
		DD nextValFn = DD.zero;

		//bestObsStrat = Util.remove(bestObsStrat, -1);		
		Vector unique = Util.unique(Util.remove(bestObsStrat, -1));

		for (int i = 0; i < unique.size(); i++)
		{
			int alphaId = (Integer) unique.get(i);
			DD obsDd = DD.zero;
			for (int obsId = 0; obsId < bestObsStrat.length; obsId++)
			{
				if (alphaId == bestObsStrat[obsId])
				{
					int[] obsConfig = statedecode(obsId + 1, p.nObsVars, p.obsVarsArity);
					int[][] merge = new int[2][];
					merge[0] = p.obsVarsPrimeId;
					merge[1] = obsConfig;
					obsDd = OP.add(obsDd, Config.convert2dd(merge));
				}
			}
			DD[] mergeDD = OP.concat(DDleaf.myNew(p.discFact), obsDd);
			mergeDD = OP.concat(mergeDD, primedV[alphaId]);

			nextValFn = OP.add(nextValFn, OP.multN(mergeDD));
		}
		DD[] mergeDD = OP.concat(p.actions[bestActId].transFn, p.actions[bestActId].obsFn);
		mergeDD = OP.concat(mergeDD, nextValFn);
		int[] vars = Util.concat(p.stateVarsPrimeId, p.obsVarsPrimeId);
		newAlpha = OP.addMultVarElim(mergeDD, vars);

		DD[] mergeNewAlpha = OP.concat(p.actions[bestActId].rewFn, newAlpha);
		newAlpha = OP.addN(mergeNewAlpha);
		//System.out.println("bestValue"+bestValue);
		bestValue = OP.factoredExpectationSparseNoMem(belState, newAlpha);
		//System.out.println("bestValue"+bestValue);

		result.add(newAlpha);
		result.add(bestValue);
		result.add(bestActId);

		return result;
	}

	/*public ArrayList dd_oneStepNZPrimeBelStates(SymbolicProblem p, DD[] belState, boolean normalize, double smallestProb ) {
		ArrayList result = new ArrayList(2);
		
		
		Vector<DD[][]> nextBelStates = new Vector(p.nActions);
		DD[] marginals = new DD[p.nStateVars+1];
		int[][] obsConfig = new int[p.nObservations][];
		for(int obsId = 0;obsId<p.nObservations;obsId++) {
			obsConfig[obsId] = statedecode(obsId+1,p.nObsVars,p.obsVarsArity); 
		}
		
		Global.newHashtables();
		Vector[] nzObsIds = new Vector[p.nActions];
		for(int actId =0;actId<p.nActions;actId++) {
			// find non-zero probability observations
			DD[] dds;
			dds = OP.concat(belState, p.actions[actId].transFn);
			dds = OP.concat(dds, p.actions[actId].obsFn);
			DD dd_obsProbs = OP.addMultVarElim(dds, Util.concat(p.stateVarsId, p.stateVarsPrimeId));
			double[] obsProbs = OP.convert2array(dd_obsProbs, p.obsVarsPrimeId);
			nzObsIds[actId] = Util.findGreater(obsProbs, smallestProb);
			
			// compute marginals
			if(!nzObsIds[actId].isEmpty()) {
				DD[][] temp = new DD[nzObsIds[actId].size()][];
				nextBelStates.add(temp);
				
				for(int varId=0;varId<p.nStateVars;varId++) {
					marginals[varId] = OP.addMultVarElim(dds,Util.concat(p.stateVarsId, Util.elimOneVar(p.stateVarsPrimeId, varId)));
				    marginals[varId] = OP.approximate(marginals[varId],1e-6, dzero());
				}
				marginals[p.nStateVars] = OP.addout(marginals[0],p.stateVarsPrimeId[0]);
			    if(normalize) {
			      DD normalizationFactor = OP.replace(marginals[p.nStateVars],0,1);
			      for(int varId = 0; varId<p.nStateVars; varId++) {
			        marginals[varId] = OP.div(marginals[varId],normalizationFactor);
			      }
			    }
			    for(int obsPtr = 0; obsPtr<nzObsIds[actId].size();obsPtr++) {
			      int obsId = (Integer)nzObsIds[actId].get(obsPtr);
			      int[][] config = new int[2][p.nObsVars];
			      config[0] = p.obsVarsPrimeId;
			      config[1] = obsConfig[obsId];			      
			      DD[][] t_next = (DD[][])nextBelStates.get(actId);
			      t_next[obsPtr] = OP.restrictN(marginals,config);
			      nextBelStates.set(actId, t_next);
			    }
			}
		}
		result.add(nextBelStates);
		result.add(nzObsIds);
		
		return result;
	}*/

	public ArrayList dd_oneStepNZPrimeBelStates(DD[] belState, boolean normalize,
			double smallestProb)
	{
		ArrayList result = new ArrayList(2);

		Vector<DD[][]> nextBelStates = new Vector(p.nActions);
		DD[] marginals = new DD[p.nStateVars + 1];
		int[][] obsConfig = new int[p.nObservations][];
		for (int obsId = 0; obsId < p.nObservations; obsId++)
		{
			obsConfig[obsId] = statedecode(obsId + 1, p.nObsVars, p.obsVarsArity);
		}
		double[] dzero = new double[1];
		dzero[0] = 0.0;

		Global.newHashtables();
		Vector[] nzObsIds = new Vector[p.nActions];
		int[][] config = new int[2][p.nObsVars];
		java.util.Vector<DD> list = new java.util.Vector();
		for (int actId = 0; actId < p.nActions; actId++)
		{
			// find non-zero probability observations
			list.clear();
			Common.addList(list, belState);
			Common.addList(list, p.actions[actId].transFn);
			Common.addList(list, p.actions[actId].obsFn);

			DD[] dds = Common.convert2array(list);

			DD dd_obsProbs = OP.addMultVarElim(dds, p.stateVarsIds);
			double[] obsProbs = OP.convert2array(dd_obsProbs, p.obsVarsPrimeId);
			nzObsIds[actId] = Util.findGreater(obsProbs, smallestProb);

			// compute marginals
			if (!nzObsIds[actId].isEmpty())
			{
				DD[][] temp = new DD[nzObsIds[actId].size()][];
				nextBelStates.add(temp);

				for (int varId = 0; varId < p.nStateVars; varId++)
				{
					marginals[varId] =
							OP.addMultVarElim(
									dds,
									Util.concat(p.stateVarsId,
											Util.elimOneVar(p.stateVarsPrimeId, varId)));
					marginals[varId] = OP.approximate(marginals[varId], 1e-6, dzero);
				}
				marginals[p.nStateVars] = OP.addout(marginals[0], p.stateVarsPrimeId[0]);

				if (normalize)
				{
					DD normalizationFactor = OP.replace(marginals[p.nStateVars], 0, 1);
					for (int varId = 0; varId < p.nStateVars; varId++)
					{
						marginals[varId] = OP.div(marginals[varId], normalizationFactor);
					}
				}
				for (int obsPtr = 0; obsPtr < nzObsIds[actId].size(); obsPtr++)
				{
					int obsId = (Integer) nzObsIds[actId].get(obsPtr);
					config[0] = p.obsVarsPrimeId;
					config[1] = obsConfig[obsId];
					DD[][] t_next = (DD[][]) nextBelStates.get(actId);
					t_next[obsPtr] = OP.restrictN(marginals, config);
					nextBelStates.set(actId, t_next);
				}
			}
		}
		result.add(nextBelStates);
		result.add(nzObsIds);

		return result;
	}

	// MODIFIED 11.09: Redundant function
	public ArrayList trim(Vector<Alpha> alphaVectors, double[] newPBV, double minImprovement)
	{
		ArrayList result = new ArrayList(3);

		int nVectors = alphaVectors.size();
		if (nVectors <= 1)
		{
			result.add(alphaVectors);
			result.add(newPBV);
			return result;
		}

		//Vector improvement = new Vector(nVectors);

		// Before and after adding alphavector,
		// If improvement is small, trim the vector
		Iterator<Alpha> it = alphaVectors.iterator();
		//for(int alphaId=0; alphaId<nVectors; alphaId++) {
		while (it.hasNext())
		{
			Alpha curAlpha = it.next();
			double[] valuesExceptAnAlpha = new double[beliefSize];

			// # of row = # of beliefs
			//System.out.println("--values--");
			for (int bpId = 0; bpId < beliefSize; bpId++)
			{
				double max = Double.NEGATIVE_INFINITY;
				for (int targetId = 0; targetId < alphaVectors.size(); targetId++)
				{
					Alpha targetAlpha = alphaVectors.get(targetId);
					//System.out.print(targetAlpha.pbv[bpId]+" ");
					if (curAlpha != targetAlpha)
					{
						double valueAlphaAtBelief = targetAlpha.pbv[bpId];
						//max = Math.max(max, ((double[])pointBasedValues.get(other_col))[row]);
						if (max < valueAlphaAtBelief)
						{
							max = valueAlphaAtBelief;
						}
					}
				}
				//System.out.println();
				valuesExceptAnAlpha[bpId] = newPBV[bpId] - max;
			}
			//improvement.add(Util.max(column));
			double improvement = Util.max(valuesExceptAnAlpha);

			// min improvement�?
			if (improvement < minImprovement)
			{
				it.remove();
			}

		}
		for (int bpId = 0; bpId < beliefSize; bpId++)
		{
			double max = Double.NEGATIVE_INFINITY;
			for (int targetId = 0; targetId < alphaVectors.size(); targetId++)
			{
				Alpha targetAlpha = alphaVectors.get(targetId);
				double valueAlphaAtBelief = targetAlpha.pbv[bpId];
				if (max < valueAlphaAtBelief)
				{
					max = valueAlphaAtBelief;
				}
			}
			newPBV[bpId] = max;
		}

		// current alpha vector set size should be under maxsize
		/*if (nVectors > maxSize) {
			int argmin = Util.argmin(improvement);
			alphaVectors.remove(argmin);
			pointBasedValues.remove(argmin);
			actions.remove(argmin);			
		}*/
		result.add(alphaVectors);
		result.add(newPBV);

		return result;
	}

	public DD[] Vector2Array(Vector<Alpha> a)
	{
		DD[] result = new DD[a.size()];
		for (int i = 0; i < a.size(); i++)
			result[i] = a.get(i).v;
		return result;
	}

	public int[] statedecode(int statenum, int n, int[] bases)
	{
		int[] stateVec = new int[n];
		for (int i = 0; i < n; i++)
			stateVec[i] = 0;

		if (statenum == 1)
		{
			for (int i = 0; i < n; i++)
				stateVec[i]++;
			return stateVec;
		}
		statenum--;
		int res = statenum;
		for (int i = 0; i < n; i++)
		{
			if (res == 1)
			{
				stateVec[i] = 1;
				break;
			}
			int remd = res % bases[i];
			res = (int) Math.floor((double) res / bases[i]);
			stateVec[i] = remd;
		}
		for (int i = 0; i < n; i++)
			stateVec[i]++;

		return stateVec;
	}
}
