package kr.ac.kaist.pomdp.utils;

import java.util.Arrays;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;

import no.uib.cipr.matrix.DenseVector;
import no.uib.cipr.matrix.Matrix;
import no.uib.cipr.matrix.Vector;
import no.uib.cipr.matrix.VectorEntry;
import no.uib.cipr.matrix.sparse.SparseVector;

public class Util
{
	public static Vector sort(Vector a)
	{
		double[] temp = new double[a.size()];
		for (int i = 0; i < temp.length; i++)
			temp[i] = a.get(i);
		Arrays.sort(temp);
		Vector res = new DenseVector(temp);
		return res;
	}

	public static double distance(Vector a, Vector b)
	{
		double dist = 0.0;
		for (int i = 0; i < a.size(); i++)
		{
			if (a.get(i) != 0 || b.get(i) != 0)
				dist += Math.abs(a.get(i) - b.get(i));
		}
		return dist;
	}

	public static int argmin(java.util.Vector p)
	{
		int index = -1;
		double min = Double.POSITIVE_INFINITY;
		for (int i = 0; i < p.size(); i++)
		{
			if ((Double) p.get(i) < min)
			{
				min = (Double) p.get(i);
				index = i;
			}
		}
		return index;
	}

	public static int[] argmax(Map<int[], Double> map)
	{
		int[] index = null;
		double max = Double.NEGATIVE_INFINITY;

		Iterator<Entry<int[], Double>> iter = map.entrySet().iterator();
		while (iter.hasNext())
		{
			Entry<int[], Double> entry = iter.next();
			if (entry.getValue() > max)
			{
				max = entry.getValue();
				index = entry.getKey();
			}
		}
		return index;
	}

	public static int argmax(Vector p)
	{
		int index = -1;
		double max = Double.NEGATIVE_INFINITY;

		for (int i = 0; i < p.size(); i++)
		{
			if (p.get(i) > max)
			{
				max = p.get(i);
				index = i;
			}
		}
		return index;
	}

	public static int argmax(double[] p)
	{
		int index = -1;
		double max = Double.NEGATIVE_INFINITY;
		for (int i = 0; i < p.length; i++)
		{
			if (p[i] > max)
			{
				max = p[i];
				index = i;
			}
		}
		return index;
	}

	public static int sample(Vector p, Random rand)
	{
		double r = rand.nextDouble();
		double cumsum = 0;
		int i = 0;

		for (i = 0; i < p.size(); i++)
		{
			if (cumsum + p.get(i) > r)
				break;
			else
				cumsum += p.get(i);
		}
		return i;
	}

	public static Vector dotMultiply(Vector a, Vector b, Vector result)
	{
		result.zero();
		for (int i = 0; i < result.size(); i++)
		{
			double val = a.get(i) * b.get(i);
			if (val != 0)
				result.set(i, val);
		}
		return result;
	}

	public static double row_sum(Vector a)
	{
		double result = 0.0;
		if (a instanceof SparseVector)
		{
			Iterator<VectorEntry> it = ((SparseVector) a).iterator();
			while (it.hasNext())
			{
				VectorEntry ve = it.next();
				result += ve.get();
			}
		}
		else
		{
			for (int i = 0; i < a.size(); i++)
				result += a.get(i);
		}
		return result;
	}

	public static double sum(double[] a)
	{
		double result = 0.0;
		for (int i = 0; i < a.length; i++)
			result += a[i];
		return result;
	}

	public static Vector normalize(Vector a)
	{
		double sum = 0.0;
		sum = row_sum(a);

		if (a instanceof SparseVector)
		{
			Iterator<VectorEntry> it = ((SparseVector) a).iterator();
			while (it.hasNext())
			{
				VectorEntry ve = it.next();
				ve.set(ve.get() / sum);
			}
		}
		else
		{
			for (int i = 0; i < a.size(); i++)
			{
				if (a.get(i) != 0)
					a.set(i, a.get(i) / sum);
			}
		}
		return a;
	}

	public static double min(Matrix[] m)
	{
		double min = Double.POSITIVE_INFINITY;
		for (int i = 0; i < m.length; i++)
			for (int j = 0; j < m[i].numRows(); j++)
				for (int k = 0; k < m[i].numColumns(); k++)
					min = Math.min(min, m[i].get(j, k));

		return min;
	}

	public static double min(java.util.Vector m)
	{
		double min = Double.POSITIVE_INFINITY;
		for (int i = 0; i < m.size(); i++)
			min = Math.min(min, (Double) m.get(i));

		return min;
	}

	public static double min(Vector m)
	{
		double min = Double.POSITIVE_INFINITY;
		for (int i = 0; i < m.size(); i++)
			min = Math.min(min, m.get(i));

		return min;
	}

	public static double min(double[] m)
	{
		double min = Double.POSITIVE_INFINITY;
		for (int i = 0; i < m.length; i++)
			min = Math.min(min, m[i]);

		return min;
	}

	public static int min(int[] m)
	{
		int min = m[0];
		for (int i = 1; i < m.length; i++)
			min = Math.min(min, m[i]);

		return min;
	}

	public static double max(double[] m)
	{
		double max = Double.NEGATIVE_INFINITY;
		for (int i = 0; i < m.length; i++)
			max = Math.max(max, m[i]);

		return max;
	}

	public static double max(java.util.Vector m)
	{
		double max = Double.NEGATIVE_INFINITY;
		for (int i = 0; i < m.size(); i++)
			max = Math.max(max, (Double) m.get(i));

		return max;
	}

	public static double max(Vector m)
	{
		double max = Double.NEGATIVE_INFINITY;
		for (int i = 0; i < m.size(); i++)
			max = Math.max(max, m.get(i));

		return max;
	}

	public static double maxAbs(double[] m)
	{
		double max = Double.MIN_VALUE;
		for (int i = 0; i < m.length; i++)
			max = Math.max(max, Math.abs(m[i]));

		return max;
	}

	public static double maxAbs(Vector m)
	{
		double max = Double.MIN_VALUE;
		for (int i = 0; i < m.size(); i++)
			max = Math.max(max, Math.abs(m.get(i)));

		return max;
	}

	public static int[] concat(int[] a, int[] b)
	{
		int[] result = new int[a.length + b.length];
		int i_res = 0;
		for (int i = 0; i < a.length; i++)
		{
			result[i_res] = a[i];
			i_res++;
		}
		for (int i = 0; i < b.length; i++)
		{
			result[i_res] = b[i];
			i_res++;
		}
		return result;
	}

	// [a , b];
	public static int[] concat(int[] b, int a)
	{
		int[] result = new int[1 + b.length];
		int i_res = 0;
		for (int i = 0; i < b.length; i++)
		{
			result[i_res] = b[i];
			i_res++;
		}
		result[b.length] = a;
		return result;
	}

	// [a , b];
	public static java.util.Vector[] concat(java.util.Vector[] b, double[] a)
	{
		for (int i = 0; i < a.length; i++)
		{
			b[i].add(a[i]);
		}
		return b;
	}

	public static int[][] concat(int[][] a, int[][] b)
	{
		int[][] result = new int[a.length][];
		for (int i = 0; i < result.length; i++)
		{
			int nCol = a[i].length + b[i].length;
			result[i] = new int[nCol];
			for (int j = 0; j < a[i].length; j++)
			{
				result[i][j] = a[i][j];
			}
			for (int j = 0; j < b[i].length; j++)
			{
				result[i][j + a[i].length] = b[i][j];
			}
		}
		return result;
	}

	public static java.util.Vector findGreater(double[] list, double x)
	{
		java.util.Vector result = new java.util.Vector(list.length);
		for (int i = 0; i < list.length; i++)
		{
			if (list[i] > x)
				result.add(i);
		}
		return result;
	}

	public static int[] elimOneVar(int[] list, int elimindex)
	{
		int[] vars = new int[list.length - 1];
		int i_vars = 0;

		for (int i = 0; i < elimindex; i++)
		{
			vars[i_vars] = list[i];
			i_vars++;
		}
		for (int i = elimindex + 1; i < list.length; i++)
		{
			vars[i_vars] = list[i];
			i_vars++;
		}
		return vars;
	}

	public static java.util.Vector unique(int[] a)
	{
		java.util.Vector result = new java.util.Vector(a.length);
		for (int i = 0; i < a.length; i++)
		{

			if (result.isEmpty())
				result.add(a[i]);
			else
			{
				if (!result.contains(a[i]))
					result.add(a[i]);
			}

		}
		return result;
	}

	public static void print(int[][] a)
	{
		for (int i = 0; i < a.length; i++)
		{
			for (int j = 0; j < a[i].length; j++)
			{
				System.out.print(a[i][j] + " ");
			}
			System.out.println();
		}
	}

	public static void print(double[][] a)
	{
		for (int i = 0; i < a.length; i++)
		{
			for (int j = 0; j < a[i].length; j++)
			{
				System.out.print(a[i][j] + " ");
			}
			System.out.println();
		}
	}

	public static void print(java.util.Vector[] a)
	{
		for (int i = 0; i < a.length; i++)
		{
			for (int j = 0; j < a[i].size(); j++)
			{
				System.out.print(a[i].get(j) + " ");
			}
			System.out.println();
		}
	}

	/**
	 * remove elements whose value is k from list
	 * 
	 * @param list
	 * 			input list
	 * @param k
	 * 			value to remove
	 * @return new list which doesn't contain value k
	 */
	public static int[] remove(int[] list, int k)
	{
		java.util.Vector temp = new java.util.Vector(list.length);
		for (int i = 0; i < list.length; i++)
		{
			if (list[i] != k)
				temp.add(list[i]);
		}
		int[] result = new int[temp.size()];
		for (int i = 0; i < temp.size(); i++)
			result[i] = (Integer) temp.get(i);
		return result;
	}

	public static Vector sub(Vector a, Vector b)
	{
		Vector res = new DenseVector(a.size());
		for (int i = 0; i < a.size(); i++)
			res.set(i, a.get(i) - b.get(i));
		return res;
	}

	public static java.util.Vector intersect(java.util.Vector a, java.util.Vector b)
	{
		java.util.Vector result = new java.util.Vector();

		Iterator it = b.iterator();
		while (it.hasNext())
		{
			Object tmp = it.next();
			if (a.contains(tmp))
				result.add(tmp);
		}

		return result;
	}
}
