package kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011;

import java.io.Serializable;

import kr.ac.kaist.pomdp.symbolic.perseus.DD;
import kr.ac.kaist.pomdp.symbolic.perseus.MySet;

/**
 * QEntry class used in Node class
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class QEntry implements Serializable {
	private static final long serialVersionUID = 4624466331380236878L;

	public double immediateReward;

	public EdgeDd outcomes;
	// public HashMap<int[], Edge> outcomes; // <observation config, edge>
	// hashmap (full version)
	public DD outcomesDd;
	public DD obsProbs;

	public double lbVal;
	public double ubVal;
	public Alpha lbAlpha;

	/**
	 * QEntry constructor
	 * 
	 * @param p
	 *            input SymbolicProblem
	 */
	public QEntry() {
		// outcomes = new HashMap<int[], Edge>();
		outcomes = null;
	}

	/**
	 * get edge corresponding to input observation id
	 * 
	 * @param obsId
	 *            input observation id
	 * @return corresponding edge
	 */
	public Edge getEdge(int[][] key) {
		return restrict(outcomes, key).getEdge();
	}

	public EdgeDd restrict(EdgeDd dd, int[][] config) {
		// should optimize restrict to terminate early by
		// exploiting variable ordering

		// it's a leaf
		if (dd.getVar() == 0) {
			return dd;
		}

		// root is variable that must be restricted
		int index = MySet.find(config[0], dd.getVar());
		if (index >= 0) {
			return restrict(dd.getChildren()[config[1][index] - 1], config);
		}

		// have to restrict all children recursively
		EdgeDd children[] = new EdgeDd[dd.getChildren().length];
		for (int i = 0; i < dd.getChildren().length; i++) {
			children[i] = restrict(dd.getChildren()[i], config);
		}
		return EdgeDdNode.myNew(dd.getVar(), children);
	}
}
